/*
 * create.c -- An HP-2000 Timeshared BASIC System tape manipulation program.
 *
 * Copyright (c) Scott G. Taylor
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * SCOTT G. TAYLOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Scott G. Taylor shall not
 * be used in advertising or otherwise to promote the sale, use or other dealings
 * in this Software without prior written authorization from Scott G. Taylor.
 */

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "config.h"

#include "tape.h"
#include "hptsb.h"
#include "decl.h"

#define		MAX_TSB_TOKEN_PROGRAM_SIZE	131072 /* Be liberal here, but probably 65536 is good */

int		ifd;
struct stat	isb;
uint32_t	ibytes;
uint32_t	rbytes;
uint32_t	verrors = 0;

void
tsb_create(void)
{
	uint32_t	i;
	struct tapeinfo tio;
	time_t          clock;
	struct tm       *ltm;
	uint32_t        clock_year,
			clock_year_hours,
			clock_hours_tenths;

	/*
	 * Validate input files.   Do as much checking as possible to ensure
	 * that this is a properly formatted and tokenized TSB file.
	 */

	for (i = 0; i < nentries; i++) {
		Dprintf (DEBUG_HEADERS, "MAIN: Validating entry %d: %s\n",
				i + 1, rentry[i]);
		ifd = open ((char *) rentry[i], O_RDONLY);
		if (ifd < 0) {
			eprintf ("ERROR: Can't open file '%s': %s\n",
					rentry[i], strerror(errno));
			verrors++;
			continue;
		}

		ibytes = fstat (ifd, &isb);
		if (ibytes < 0) {
			eprintf ("ERROR: Can't stat file '%s': %s\n",
					rentry[i], strerror(errno));
			verrors++;
			close (ifd);
			continue;
		}

		ibytes = read (ifd, buf, 24);
		if (ibytes != 24) {
			eprintf ("ERROR: Can't read TSB header of file '%s'.\n",
					rentry[i]);
			verrors++;
			close(ifd);
			continue;
		}
		if (!S_ISREG(isb.st_mode)) {
			eprintf ("ERROR: Not a regular file: %s\n",
					rentry[i]);
			verrors++;
			close(ifd);
			continue;
		}
		rbytes = isb.st_size - 24;

		Pentry.id = WTOI (buf);
		memcpy (Pentry.HPfilnam, buf + 2, 6);
		Pentry.startloc = WTOI (buf + 8);
		Pentry.refdate = WTOI (buf + 10);
		Pentry.moddate = WTOI (buf + 12);
		Pentry.status = WTOI (buf + 14);
		Pentry.daddr = WTOI (buf + 16);
		Pentry.ddev = WTOI (buf + 18);
		Pentry.linfo = WTOI (buf + 20);
		Pentry.len = WTOI (buf + 22);
		Pentry.type = 0;

		if (Pentry.HPfilnam[4] & 0x80) {
			Pentry.type = ENTRY_TYPE_SET_CSAVE;
			Pentry.HPfilnam[4] &= 0x7f;
			Dprintf (DEBUG_FILEREAD, "1-type CSAVE\n");
		}
		if (Pentry.HPfilnam[2] & 0x80) {
			Pentry.type = ENTRY_TYPE_SET_FILE;
			Pentry.HPfilnam[2] &= 0x7f;
			Dprintf (DEBUG_FILEREAD, "1-type FILE\n");
		}
		if (Pentry.HPfilnam[0] & 0x80) {
			Pentry.type = ENTRY_TYPE_SET_ASCII;
			Pentry.HPfilnam[0] &= 0x7f;
			Dprintf (DEBUG_FILEREAD, "1-type ASCII\n");
		}
		if (Pentry.type == 0)
			Dprintf (DEBUG_FILEREAD, "1-type PROGRAM\n");
		if (ENTRY_TYPE_PROGRAM(Pentry.type)) {
			Pentry.filesize = (65536 - Pentry.len) * 2;
			Pentry.block_count = (Pentry.filesize + 2047) / 2048;
		} else if (ENTRY_TYPE_FILE(Pentry.type)) { 
			Pentry.filesize = (Pentry.len * 256) * 2;
			Pentry.block_count = (Pentry.filesize + (Pentry.startloc - 1)) / 2048;
		} else if (ENTRY_TYPE_ASCII(Pentry.type)) {
			Pentry.filesize = (Pentry.len * 256) * 2;
			Pentry.block_count = Pentry.filesize / 2048;
		}
		if (ibytes < 24) {
			eprintf ("ERROR: Invalid TSB input file: %s\n",
					rentry[i]);
			verrors++;
		} else if (!validate_id (WTOI(buf))) {
			eprintf ("ERROR: Invalid User ID in input file: %s\n",
					rentry[i]);
			verrors++;
		} else if (!validate_filnam((uint8_t *) buf + 2, MACH_TYPE_SET_ACCESSB)) {
			eprintf ("ERROR: Invalid filnam in input file '%s': %-6.6s\n",
					rentry[i], buf + 2);
			verrors++;
		} else if (Pentry.filesize < 0 || Pentry.filesize > (32767 * 512)) {
			eprintf ("ERROR: Invalid file size in input file '%s'.\n",
					rentry[i]);
			verrors++;
		} else if ((Pentry.refdate & 0x1ff) > 366 || (Pentry.refdate >> 9) > 99 ||
				(Pentry.refdate & 0x1ff) == 0) {
			eprintf ("ERROR: Invalid REFDATE field in file header of '%s'.\n",
					rentry[i]);
			verrors++;
		} else if (Pentry.moddate > 8784) {
			eprintf ("ERROR: Invalid MODDATE field in file header of '%s'.\n",
					rentry[i]);
				verrors++;
		} else if (rbytes < Pentry.filesize) {
			eprintf ("ERROR: Corrupt file '%s' -- too short -- Should be %d+24 bytes, got %d+24: %s\n",
					rentry[i], Pentry.filesize, rbytes,
					strerror(errno));
			verrors++;
		} else {
			eprintf ("USER ID: %c%03d  FILNAM: %-6.6s  FILSIZ=%d\t%s\n",
					Pentry.id / 1024 + '@', Pentry.id % 1024,
					Pentry.HPfilnam, Pentry.filesize,rentry[i]);
		}
		close (ifd);
	}

	if (verrors) {
		eprintf ("ERROR: %d Invalid TSB file%s (of %d) specified.  Can't continue\n",
				verrors, verrors == 1 ? "" : "s", nentries);
		exit (-99);
	}

/*debug = (2 << DEBUG_HEADERS);*/
	Dprintf (DEBUG_HEADERS, "MAIN: Writing tape %s\n", infilename[0]);

	tio.maxbsize = MAX_TAPE_RECLEN;
	tio.debug = (debug & (2 << DEBUG_TAPE)) != 0;

	tape_write_open (&tio, (uint8_t *) infilename[0]);
	Dprintf (DEBUG_HEADERS, "MAIN: Tape opened, tio.fd=%d\n",
			tio.fd);

	clock = time(0L);
	ltm = localtime (&clock);
	clock_year = ltm->tm_year % 100;
	clock_year_hours = (ltm->tm_yday - 1) * 24 + ltm->tm_hour;
	clock_hours_tenths = ltm->tm_min * 600 + ltm->tm_sec * 10;

	Pentry.header[0] = ITOWH (65536 - 10);		/* 10 word record */
	Pentry.header[1] = ITOWL (65536 - 10);
	Pentry.header[2] = 'L';				/* Label "LBTS" */
	Pentry.header[3] = 'B';
	Pentry.header[4] = 'T';
	Pentry.header[5] = 'S';
	Pentry.header[6] = 0;				/* Unused */
	Pentry.header[7] = 0;				/* Unused */
	Pentry.header[8] = ITOWH (1);			/* Tape Vol number */
	Pentry.header[9] = ITOWL (1);
	Pentry.header[10] = ITOWH (clock_year);		/* 2-digit Year */
	Pentry.header[11] = ITOWL (clock_year);
	Pentry.header[12] = ITOWH (clock_year_hours);	/* Hour of year */
	Pentry.header[13] = ITOWL (clock_year_hours);
	Pentry.header[14] = ITOWH (clock_hours_tenths);	/* Tenths of hour */
	Pentry.header[15] = ITOWL (clock_hours_tenths);
	Pentry.header[16] = ITOWH (6000);		/* System Level 6000 */
	Pentry.header[17] = ITOWL (6000);
	Pentry.header[18] = ITOWH (1500);		/* Feature Level 1500 */
	Pentry.header[19] = ITOWL (1500);

	Dprintf (DEBUG_HEADERS, "MAIN: Writing tape header label\n");
	tape_write (&tio, Pentry.header, 20);
	tape_write_eof (&tio);

	for (i = 0; i < nentries; i++) {
		Dprintf (DEBUG_HEADERS, "MAIN: Writing entry %d to tape: %s\n",
				i + 1, rentry[i]);
		ifd = open ((char *) rentry[i], O_RDONLY);
		Dprintf (DEBUG_HEADERS, "MAIN: Entry opened, fd=%d: %s\n",
				ifd, rentry[i]);
		if (ifd < 0)
			errorexit (-50, "ERROR: Can't open file '%s': %s\n",
					rentry[i], strerror(errno));
		ibytes = read (ifd, buf, 2048 + 24);
		Dprintf (DEBUG_HEADERS, "MAIN: Writing first entry record of %d bytes\n",
				ibytes);
		/* First record is the 24-byte directory plus up to 2048 bytes data. */
		tape_write (&tio, (uint8_t *) buf, ibytes);
		while ((ibytes = read (ifd, buf, 2048)) > 0) {
			/*
			 * Consecutive tape records are max 2048 bytes, minimum 24 bytes,
			 * and even byte counts.
			 */
			tape_write (&tio, (uint8_t *) buf, ibytes > 24 ? ibytes + (ibytes % 2) : 24);
			Dprintf (DEBUG_HEADERS, "MAIN: Writing next entry record of %d bytes\n",
				ibytes);
		}
		tape_write_eof (&tio);
		close (ifd);
	}

	Dprintf (DEBUG_HEADERS, "MAIN: Closing tape file.\n");
	tape_close (&tio);	
}

void
tsb_tokenize (void)
{
	int xfd;
	uint32_t	i;
#ifdef OLD
	struct tapeinfo tio;
	time_t          clock;
	struct tm       *ltm;
	uint32_t        clock_year,
			clock_year_hours,
			clock_hours_tenths;
#endif
	uint8_t		*kdata;
	struct Program_Entry Pentry1;
	struct Program_Entry Pentry2;
	struct Program_Space Pspace;

	FILE		*Ffd;
	uint32_t	lineno;
	struct line_table *ltab, *ltab2;
	uint8_t		line[512];
	uint8_t		*Pblock;


	uint8_t		newfilename[MAXPATHLEN];
	/*
	 * Validate input files.   Do as much checking as possible to ensure
	 * that this is a properly formatted and tokenized TSB file.
	 */

/*debug=0xffffffff;*/
	for (i = 0; i < nentries; i++) {
		Dprintf (DEBUG_HEADERS, "MAIN: Validating entry %d: %s\n",
				i + 1, rentry[i]);

		sprintf ((char *) newfilename, "/tiny/hptapes/patti.newtoken/%s", rentry[i]);

		if (!strcmp((char *) rentry[i], "-")) {
			ifd = open ("/dev/tty", O_RDONLY);
		} else {
			ifd = open ((char *) rentry[i], O_RDONLY);
		}
		if (ifd < 0) {
			eprintf ("ERROR: Can't open file '%s': %s\n",
					rentry[i], strerror(errno));
			verrors++;
			continue;
		}

		ibytes = fstat (ifd, &isb);
		if (ibytes < 0) {
			eprintf ("ERROR: Can't stat file '%s': %s\n",
					rentry[i], strerror(errno));
			verrors++;
			close (ifd);
			continue;
		}

		if (strcmp ((char *) rentry[i], "-") && !S_ISREG(isb.st_mode)) {
			eprintf ("ERROR: Not a regular file: %s\n",
					rentry[i]);
			verrors++;
			close(ifd);
			continue;
		}


		Pentry1.block = malloc (MAX_TSB_TOKEN_PROGRAM_SIZE); /* Tokenized bytes go here */
		Pentry1.block_size = 0;		/* Use this to keep track of bytes loaded into block */
		Pblock = Pentry1.block;
		tape_mach_type = MACH_TYPE_SET_ACCESSB; /* Tokenization type is for 2000/ACCESS 1B */
		Pentry1.type = 0;		/* Just a regular tokenized program */
		Pentry1.startloc = 02000;	/* Always 02000 for 2000/ACCESS */

		/* Load up Pentry.block */

		Ffd = fdopen (ifd, "r");
		ltab = &Pentry1.cline;
		lineno = 0;

eprintf ("READING INPUT FILE: %s\n", rentry[i]);
		while (fgets ((char *) line, 511, Ffd) != NULL) {
			
			lineno ++;
			line[strlen((char *) line) - 1] = 0;
			ltab->len = strlen((char *) line);
			ltab->data = malloc (ltab->len + 1);
			memcpy (ltab->data, line, ltab->len + 1);

			ltab->block_linestart = Pblock;
			if (tokenize_line (ltab->data, &Pspace, NULL)) {
				eprintf ("ERROR: tokenizing line %d\n", lineno);
				eprintf ("ERROR: %s\n", ltab->data);
				eprintf ("ERROR: ");
				for (kdata = ltab->data; kdata < Pspace.error_pos; kdata++)
					eprintf (" ");
				eprintf ("^\n");
				eprintf ("ERROR: %s\n", Pspace.error_text);
			} else {
				Pentry1.block_size += Pspace.statement_len * 2;
				memcpy (Pblock, Pspace.statement, Pspace.statement_len * 2);
				Pblock += Pspace.statement_len * 2;
			}

			ltab->next = malloc (sizeof (struct line_table));
			ltab = ltab->next;
			ltab->data = NULL;
			ltab->next = NULL;
			ltab->len = 0;
		}
		
		eprintf ("%d lines read.\n", lineno);
		fclose (Ffd);

		/* Set up Pentry2 */

		tape_mach_type = MACH_TYPE_SET_ACCESSB; /* Tokenization type is for 2000/ACCESS 1B */
		Pentry2.type = 0;		/* Just a regular tokenized program */
		Pentry2.startloc = 02000;	/* Always 02000 for 2000/ACCESS */

		Pentry2.block_size = Pentry1.block_size;
		Pentry2.filesize = Pentry2.block_size;
		Pentry2.block = malloc (Pentry1.block_size);
		memcpy (Pentry2.block, Pentry1.block, Pentry2.filesize);
		Pentry2.symtabstart = Pentry2.startloc + Pentry2.filesize; /* We never CSAVE, the filesize + 02000 */
		verbose = 0x00ff;
		convert = 1;

eprintf ("Calling convert_program, block_size=%d  filesize=%d\n", Pentry2.block_size, Pentry2.filesize);

/*xfd = open ("x.x.tsbout", O_WRONLY | O_CREAT, 0666);*/
xfd = open ((char *) newfilename, O_WRONLY | O_CREAT, 0666);
write (xfd, Pentry2.block, Pentry2.block_size);
close (xfd);

		convert_program (&Pentry2);

		/* Compare the two */

		ltab = &Pentry1.cline;
		ltab2 = &Pentry2.cline;
		lineno = 0;

eprintf ("Comparing input ASCII to output ASCII\n");
		while (ltab != NULL && ltab2 != NULL && ltab->len != 0 && ltab2->len != 0) {
			lineno++;
			if (ltab->len != ltab2->len && strncmp((char *) ltab->data, (char *) ltab2->data, ltab->len)) {
				eprintf ("ERR: ASCII Comparison failed:\n");
				eprintf (" IN> %.*s\n", ltab->len, ltab->data);
				eprintf ("OUT> %.*s\n", ltab2->len, ltab2->data);
			} else {
				eprintf (" OK:\n");
				eprintf (" IN> %.*s\n", ltab->len, ltab->data);
				eprintf ("OUT> %.*s\n", ltab2->len, ltab2->data);
			}
			ltab = ltab->next;
			ltab2 = ltab2->next;
		}
				
		/* Call tsb_tokenize(); */
		ltab = &Pentry1.cline;

		if (&Pentry1.cline != NULL) { 
			struct line_table *optr;
			while (ltab != NULL) {
				if (ltab->data != NULL) {
eprintf (" IN> %.*s\n", ltab->len, ltab->data);
					nfree(ltab->data);
				}
				optr = ltab;
				ltab = ltab->next;
				if (optr != &Pentry1.cline)
					if (optr != NULL)
						nfree(optr);
			}
		}

		ltab = &Pentry2.cline;
		if (&Pentry2.cline != NULL) { 
			struct line_table *optr;
			while (ltab != NULL) {
				if (ltab->data != NULL) {
eprintf ("OUT> %.*s\n", ltab->len, ltab->data);
					nfree(ltab->data);
				}
				optr = ltab;
				ltab = ltab->next;
				if (optr != &Pentry2.cline)
					if (optr != NULL)
						nfree(optr);
			}
		}

		if (Pentry1.block != NULL)
			free(Pentry1.block);
		if (Pentry2.block != NULL)
			free(Pentry2.block);

	}

}
